@@ -11,6 +11,7 @@ from mypy.version import __version__ as mypy_version
 
 
 T = TypeVar('T')
+allowed_duplicates = ['@overload', 'Got:', 'Expected:']
 
 
 class ErrorInfo:
@@ -65,8 +66,8 @@ class ErrorInfo:
                  message: str,
                  blocker: bool,
                  only_once: bool,
-                 origin: Tuple[str, int] = None,
-                 target: str = None) -> None:
+                 origin: Optional[Tuple[str, int]] = None,
+                 target: Optional[str] = None) -> None:
         self.import_ctx = import_ctx
         self.file = file
         self.module = module
@@ -175,7 +176,9 @@ class Errors:
         file = os.path.normpath(file)
         return remove_path_prefix(file, self.ignore_prefix)
 
-    def set_file(self, file: str, module: Optional[str], ignored_lines: Set[int] = None) -> None:
+    def set_file(self, file: str,
+                 module: Optional[str],
+                 ignored_lines: Optional[Set[int]] = None) -> None:
         """Set the path and module id of the current file."""
         # The path will be simplified later, in render_messages. That way
         #  * 'file' is always a key that uniquely identifies a source file
@@ -251,9 +254,16 @@ class Errors:
         """Replace the entire import context with a new value."""
         self.import_ctx = ctx[:]
 
-    def report(self, line: int, column: int, message: str, blocker: bool = False,
-               severity: str = 'error', file: str = None, only_once: bool = False,
-               origin_line: int = None) -> None:
+    def report(self,
+               line: int,
+               column: int,
+               message: str,
+               blocker: bool = False,
+               severity: str = 'error',
+               file: Optional[str] = None,
+               only_once: bool = False,
+               origin_line: Optional[int] = None,
+               offset: int = 0) -> None:
         """Report message at the given line using the current error context.
 
         Args:
@@ -270,6 +280,8 @@ class Errors:
             type = None  # Omit type context if nested function
         if file is None:
             file = self.file
+        if offset:
+            message = " " * offset + message
         info = ErrorInfo(self.import_context(), file, self.current_module(), type,
                          self.function_or_member[-1], line, column, severity, message,
                          blocker, only_once,
@@ -471,6 +483,10 @@ class Errors:
             while (j >= 0 and errors[j][0] == errors[i][0] and
                     errors[j][1] == errors[i][1]):
                 if (errors[j][3] == errors[i][3] and
+                        # Allow duplicate notes in overload conficts reporting
+                        not (errors[i][3] == 'note' and
+                             errors[i][4].strip() in allowed_duplicates
+                             or errors[i][4].strip().startswith('def ')) and
                         errors[j][4] == errors[i][4]):  # ignore column
                     dup = True
                     break
@@ -514,7 +530,7 @@ def remove_path_prefix(path: str, prefix: str) -> str:
         return path
 
 
-def report_internal_error(err: Exception, file: str, line: int,
+def report_internal_error(err: Exception, file: Optional[str], line: int,
                           errors: Errors, options: Options) -> None:
     """Report internal error and exit.
 
@@ -529,13 +545,16 @@ def report_internal_error(err: Exception, file: str, line: int,
         print("Failed to dump errors:", repr(e), file=sys.stderr)
 
     # Compute file:line prefix for official-looking error messages.
-    if line:
-        prefix = '{}:{}'.format(file, line)
+    if file:
+        if line:
+            prefix = '{}:{}: '.format(file, line)
+        else:
+            prefix = '{}: '.format(file)
     else:
-        prefix = file
+        prefix = ''
 
     # Print "INTERNAL ERROR" message.
-    print('{}: error: INTERNAL ERROR --'.format(prefix),
+    print('{}error: INTERNAL ERROR --'.format(prefix),
           'please report a bug at https://github.com/python/mypy/issues',
           'version: {}'.format(mypy_version),
           file=sys.stderr)
