@@ -2,12 +2,16 @@
 
 import cgi
 import os.path
+import typing
 
-from typing import Dict, List, cast, Tuple, Set, Optional
+from collections import Counter
+from typing import Dict, List, cast, Tuple, Optional
 
 from mypy.traverser import TraverserVisitor
+from mypy.typeanal import collect_all_inner_types
 from mypy.types import (
-    Type, AnyType, Instance, FunctionLike, TupleType, TypeVarType, TypeQuery, CallableType
+    Type, AnyType, Instance, FunctionLike, TupleType, TypeVarType, TypeQuery, CallableType,
+    TypeOfAny
 )
 from mypy import nodes
 from mypy.nodes import (
@@ -32,28 +36,37 @@ precision_names = [
 
 
 class StatisticsVisitor(TraverserVisitor):
-    def __init__(self, inferred: bool, filename: str, typemap: Dict[Expression, Type] = None,
-                 all_nodes: bool = False) -> None:
+    def __init__(self,
+                 inferred: bool,
+                 filename: str,
+                 typemap: Optional[Dict[Expression, Type]] = None,
+                 all_nodes: bool = False,
+                 visit_untyped_defs: bool = True) -> None:
         self.inferred = inferred
         self.filename = filename
         self.typemap = typemap
         self.all_nodes = all_nodes
+        self.visit_untyped_defs = visit_untyped_defs
 
-        self.num_precise = 0
-        self.num_imprecise = 0
-        self.num_any = 0
+        self.num_precise_exprs = 0
+        self.num_imprecise_exprs = 0
+        self.num_any_exprs = 0
 
-        self.num_simple = 0
-        self.num_generic = 0
-        self.num_tuple = 0
-        self.num_function = 0
-        self.num_typevar = 0
-        self.num_complex = 0
+        self.num_simple_types = 0
+        self.num_generic_types = 0
+        self.num_tuple_types = 0
+        self.num_function_types = 0
+        self.num_typevar_types = 0
+        self.num_complex_types = 0
+        self.num_any_types = 0
 
         self.line = -1
 
         self.line_map = {}  # type: Dict[int, int]
 
+        self.type_of_any_counter = Counter()  # type: typing.Counter[TypeOfAny.TypeOfAny]
+        self.any_line_map = {}  # type: Dict[int, List[AnyType]]
+
         self.output = []  # type: List[str]
 
         TraverserVisitor.__init__(self)
@@ -79,7 +92,8 @@ class StatisticsVisitor(TraverserVisitor):
                 self.type(sig.ret_type)
             elif self.all_nodes:
                 self.record_line(self.line, TYPE_ANY)
-            super().visit_func_def(o)
+            if not o.is_dynamic() or self.visit_untyped_defs:
+                super().visit_func_def(o)
 
     def visit_class_def(self, o: ClassDef) -> None:
         # Override this method because we don't want to analyze base_type_exprs (base_type_exprs
@@ -104,7 +118,8 @@ class StatisticsVisitor(TraverserVisitor):
             return
         if o.type:
             self.type(o.type)
-        elif self.inferred:
+        elif self.inferred and not self.all_nodes:
+            # if self.all_nodes is set, lvalues will be visited later
             for lvalue in o.lvalues:
                 if isinstance(lvalue, nodes.TupleExpr):
                     items = lvalue.items
@@ -115,15 +130,7 @@ class StatisticsVisitor(TraverserVisitor):
                 for item in items:
                     if isinstance(item, RefExpr) and item.is_def:
                         if self.typemap is not None:
-                            t = self.typemap.get(item)
-                        else:
-                            t = None
-                        if t:
-                            self.type(t)
-                        else:
-                            self.log('  !! No inferred type on line %d' %
-                                     self.line)
-                            self.record_line(self.line, TYPE_ANY)
+                            self.type(self.typemap.get(item))
         super().visit_assignment_stmt(o)
 
     def visit_name_expr(self, o: NameExpr) -> None:
@@ -177,40 +184,52 @@ class StatisticsVisitor(TraverserVisitor):
             self.record_line(self.line, TYPE_UNANALYZED)
             return
 
-        if isinstance(t, AnyType) and t.special_form:
+        if isinstance(t, AnyType) and t.type_of_any == TypeOfAny.special_form:
             # This is not a real Any type, so don't collect stats for it.
             return
 
         if isinstance(t, AnyType):
             self.log('  !! Any type around line %d' % self.line)
-            self.num_any += 1
+            self.num_any_exprs += 1
             self.record_line(self.line, TYPE_ANY)
         elif ((not self.all_nodes and is_imprecise(t)) or
               (self.all_nodes and is_imprecise2(t))):
             self.log('  !! Imprecise type around line %d' % self.line)
-            self.num_imprecise += 1
+            self.num_imprecise_exprs += 1
             self.record_line(self.line, TYPE_IMPRECISE)
         else:
-            self.num_precise += 1
+            self.num_precise_exprs += 1
             self.record_line(self.line, TYPE_PRECISE)
 
-        if isinstance(t, Instance):
-            if t.args:
-                if any(is_complex(arg) for arg in t.args):
-                    self.num_complex += 1
+        for typ in collect_all_inner_types(t) + [t]:
+            if isinstance(typ, AnyType):
+                if typ.type_of_any == TypeOfAny.from_another_any:
+                    assert typ.source_any
+                    assert typ.source_any.type_of_any != TypeOfAny.from_another_any
+                    typ = typ.source_any
+                self.type_of_any_counter[typ.type_of_any] += 1
+                self.num_any_types += 1
+                if self.line in self.any_line_map:
+                    self.any_line_map[self.line].append(typ)
                 else:
-                    self.num_generic += 1
-            else:
-                self.num_simple += 1
-        elif isinstance(t, FunctionLike):
-            self.num_function += 1
-        elif isinstance(t, TupleType):
-            if any(is_complex(item) for item in t.items):
-                self.num_complex += 1
-            else:
-                self.num_tuple += 1
-        elif isinstance(t, TypeVarType):
-            self.num_typevar += 1
+                    self.any_line_map[self.line] = [typ]
+            elif isinstance(typ, Instance):
+                if typ.args:
+                    if any(is_complex(arg) for arg in typ.args):
+                        self.num_complex_types += 1
+                    else:
+                        self.num_generic_types += 1
+                else:
+                    self.num_simple_types += 1
+            elif isinstance(typ, FunctionLike):
+                self.num_function_types += 1
+            elif isinstance(typ, TupleType):
+                if any(is_complex(item) for item in typ.items):
+                    self.num_complex_types += 1
+                else:
+                    self.num_tuple_types += 1
+            elif isinstance(typ, TypeVarType):
+                self.num_typevar_types += 1
 
     def log(self, string: str) -> None:
         self.output.append(string)
@@ -221,7 +240,7 @@ class StatisticsVisitor(TraverserVisitor):
 
 
 def dump_type_stats(tree: MypyFile, path: str, inferred: bool = False,
-                    typemap: Dict[Expression, Type] = None) -> None:
+                    typemap: Optional[Dict[Expression, Type]] = None) -> None:
     if is_special_module(path):
         return
     print(path)
@@ -230,17 +249,17 @@ def dump_type_stats(tree: MypyFile, path: str, inferred: bool = False,
     for line in visitor.output:
         print(line)
     print('  ** precision **')
-    print('  precise  ', visitor.num_precise)
-    print('  imprecise', visitor.num_imprecise)
-    print('  any      ', visitor.num_any)
+    print('  precise  ', visitor.num_precise_exprs)
+    print('  imprecise', visitor.num_imprecise_exprs)
+    print('  any      ', visitor.num_any_exprs)
     print('  ** kinds **')
-    print('  simple   ', visitor.num_simple)
-    print('  generic  ', visitor.num_generic)
-    print('  function ', visitor.num_function)
-    print('  tuple    ', visitor.num_tuple)
-    print('  TypeVar  ', visitor.num_typevar)
-    print('  complex  ', visitor.num_complex)
-    print('  any      ', visitor.num_any)
+    print('  simple   ', visitor.num_simple_types)
+    print('  generic  ', visitor.num_generic_types)
+    print('  function ', visitor.num_function_types)
+    print('  tuple    ', visitor.num_tuple_types)
+    print('  TypeVar  ', visitor.num_typevar_types)
+    print('  complex  ', visitor.num_complex_types)
+    print('  any      ', visitor.num_any_types)
 
 
 def is_special_module(path: str) -> bool:
@@ -329,7 +348,8 @@ def generate_html_report(tree: MypyFile, path: str, type_map: Dict[Expression, T
             status = visitor.line_map.get(lineno, TYPE_PRECISE)
             style_map = {TYPE_PRECISE: 'white',
                          TYPE_IMPRECISE: 'yellow',
-                         TYPE_ANY: 'red'}
+                         TYPE_ANY: 'red',
+                         TYPE_UNANALYZED: 'red'}
             style = style_map[status]
             append('<span class="lineno">%4d</span>   ' % lineno +
                    '<span class="%s">%s</span>' % (style,
