@@ -5,7 +5,7 @@ from abc import abstractmethod
 from collections import OrderedDict
 from typing import (
     Any, TypeVar, Dict, List, Tuple, cast, Generic, Set, Optional, Union, Iterable, NamedTuple,
-    Callable,
+    Callable
 )
 
 import mypy.nodes
@@ -172,7 +172,7 @@ class UnboundType(Type):
 
     def __init__(self,
                  name: str,
-                 args: List[Type] = None,
+                 args: Optional[List[Type]] = None,
                  line: int = -1,
                  column: int = -1,
                  optional: bool = False,
@@ -188,6 +188,15 @@ class UnboundType(Type):
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_unbound_type(self)
 
+    def __hash__(self) -> int:
+        return hash((self.name, self.optional, tuple(self.args)))
+
+    def __eq__(self, other: object) -> bool:
+        if not isinstance(other, UnboundType):
+            return NotImplemented
+        return (self.name == other.name and self.optional == other.optional and
+                self.args == other.args)
+
     def serialize(self) -> JsonDict:
         return {'.class': 'UnboundType',
                 'name': self.name,
@@ -250,65 +259,85 @@ class TypeList(Type):
 _dummy = object()  # type: Any
 
 
+class TypeOfAny:
+    """
+    This class describes different types of Any. Each 'Any' can be of only one type at a time.
+
+    TODO: this class should be made an Enum once we drop support for python 3.3.
+    """
+    MYPY = False
+    if MYPY:
+        from typing import NewType
+        TypeOfAny = NewType('TypeOfAny', str)
+    else:
+        def TypeOfAny(x: str) -> str:
+            return x
+
+    # Was this Any type was inferred without a type annotation?
+    unannotated = TypeOfAny('unannotated')
+    # Does this Any come from an explicit type annotation?
+    explicit = TypeOfAny('explicit')
+    # Does this come from an unfollowed import? See --disallow-any=unimported option
+    from_unimported_type = TypeOfAny('from_unimported_type')
+    # Does this Any type come from omitted generics?
+    from_omitted_generics = TypeOfAny('from_omitted_generics')
+    # Does this Any come from an error?
+    from_error = TypeOfAny('from_error')
+    # Is this a type that can't be represented in mypy's type system? For instance, type of
+    # call to NewType(...)). Even though these types aren't real Anys, we treat them as such.
+    special_form = TypeOfAny('special_form')
+    # Does this Any come from interaction with another Any?
+    from_another_any = TypeOfAny('from_another_any')
+
+
 class AnyType(Type):
     """The type 'Any'."""
 
     def __init__(self,
-                 implicit: bool = False,
-                 from_unimported_type: bool = False,
-                 explicit: bool = False,
-                 from_omitted_generics: bool = False,
-                 special_form: bool = False,
+                 type_of_any: TypeOfAny.TypeOfAny,
+                 source_any: Optional['AnyType'] = None,
                  line: int = -1,
                  column: int = -1) -> None:
         super().__init__(line, column)
-        # Was this Any type was inferred without a type annotation?
-        # Note that this is not always the opposite of explicit.
-        # For instance, if "Any" comes from an unimported type,
-        # both explicit and implicit will be False
-        self.implicit = implicit
-        # Does this come from an unfollowed import? See --disallow-any=unimported option
-        self.from_unimported_type = from_unimported_type
-        # Does this Any come from an explicit type annotation?
-        self.explicit = explicit
-        # Does this type come from omitted generics?
-        self.from_omitted_generics = from_omitted_generics
-        # Is this a type that can't be represented in mypy's type system? For instance, type of
-        # call to NewType(...)). Even though these types aren't real Anys, we treat them as such.
-        self.special_form = special_form
+        self.type_of_any = type_of_any
+        # If this Any was created as a result of interacting with another 'Any', record the source
+        # and use it in reports.
+        self.source_any = source_any
+        if source_any and source_any.source_any:
+            self.source_any = source_any.source_any
+
+        # Only Anys that come from another Any can have source_any.
+        assert type_of_any != TypeOfAny.from_another_any or source_any is not None
+        # We should not have chains of Anys.
+        assert not self.source_any or self.source_any.type_of_any != TypeOfAny.from_another_any
 
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_any(self)
 
     def copy_modified(self,
-                      implicit: bool = _dummy,
-                      from_unimported_type: bool = _dummy,
-                      explicit: bool = _dummy,
-                      from_omitted_generics: bool = _dummy,
-                      special_form: bool = _dummy,
+                      type_of_any: TypeOfAny.TypeOfAny = _dummy,
+                      original_any: Optional['AnyType'] = _dummy,
                       ) -> 'AnyType':
-        if implicit is _dummy:
-            implicit = self.implicit
-        if from_unimported_type is _dummy:
-            from_unimported_type = self.from_unimported_type
-        if explicit is _dummy:
-            explicit = self.explicit
-        if from_omitted_generics is _dummy:
-            from_omitted_generics = self.from_omitted_generics
-        if special_form is _dummy:
-            special_form = self.special_form
-        return AnyType(implicit=implicit, from_unimported_type=from_unimported_type,
-                       explicit=explicit, from_omitted_generics=from_omitted_generics,
-                       special_form=special_form,
+        if type_of_any is _dummy:
+            type_of_any = self.type_of_any
+        if original_any is _dummy:
+            original_any = self.source_any
+        return AnyType(type_of_any=type_of_any, source_any=original_any,
                        line=self.line, column=self.column)
 
+    def __hash__(self) -> int:
+        return hash(AnyType)
+
+    def __eq__(self, other: object) -> bool:
+        return isinstance(other, AnyType)
+
     def serialize(self) -> JsonDict:
         return {'.class': 'AnyType'}
 
     @classmethod
     def deserialize(cls, data: JsonDict) -> 'AnyType':
         assert data['.class'] == 'AnyType'
-        return AnyType()
+        return AnyType(TypeOfAny.special_form)
 
 
 class UninhabitedType(Type):
@@ -336,6 +365,12 @@ class UninhabitedType(Type):
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_uninhabited_type(self)
 
+    def __hash__(self) -> int:
+        return hash(UninhabitedType)
+
+    def __eq__(self, other: object) -> bool:
+        return isinstance(other, UninhabitedType)
+
     def serialize(self) -> JsonDict:
         return {'.class': 'UninhabitedType',
                 'is_noreturn': self.is_noreturn}
@@ -357,6 +392,12 @@ class NoneTyp(Type):
     def __init__(self, line: int = -1, column: int = -1) -> None:
         super().__init__(line, column)
 
+    def __hash__(self) -> int:
+        return hash(NoneTyp)
+
+    def __eq__(self, other: object) -> bool:
+        return isinstance(other, NoneTyp)
+
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_none_type(self)
 
@@ -388,7 +429,7 @@ class DeletedType(Type):
 
     source = ''  # type: Optional[str]  # May be None; name that generated this value
 
-    def __init__(self, source: str = None, line: int = -1, column: int = -1) -> None:
+    def __init__(self, source: Optional[str] = None, line: int = -1, column: int = -1) -> None:
         self.source = source
         super().__init__(line, column)
 
@@ -436,6 +477,14 @@ class Instance(Type):
 
     type_ref = None  # type: str
 
+    def __hash__(self) -> int:
+        return hash((self.type, tuple(self.args)))
+
+    def __eq__(self, other: object) -> bool:
+        if not isinstance(other, Instance):
+            return NotImplemented
+        return self.type == other.type and self.args == other.args
+
     def serialize(self) -> Union[JsonDict, str]:
         assert self.type is not None
         type_ref = self.type.fullname()
@@ -498,6 +547,14 @@ class TypeVarType(Type):
         else:
             return self.upper_bound
 
+    def __hash__(self) -> int:
+        return hash(self.id)
+
+    def __eq__(self, other: object) -> bool:
+        if not isinstance(other, TypeVarType):
+            return NotImplemented
+        return self.id == other.id
+
     def serialize(self) -> JsonDict:
         assert not self.id.is_meta_var()
         return {'.class': 'TypeVarType',
@@ -588,9 +645,9 @@ class CallableType(FunctionLike):
                  arg_names: List[Optional[str]],
                  ret_type: Type,
                  fallback: Instance,
-                 name: str = None,
-                 definition: SymbolNode = None,
-                 variables: List[TypeVarDef] = None,
+                 name: Optional[str] = None,
+                 definition: Optional[SymbolNode] = None,
+                 variables: Optional[List[TypeVarDef]] = None,
                  line: int = -1,
                  column: int = -1,
                  is_ellipsis_args: bool = False,
@@ -598,7 +655,7 @@ class CallableType(FunctionLike):
                  is_classmethod_class: bool = False,
                  special_sig: Optional[str] = None,
                  from_type_type: bool = False,
-                 bound_args: List[Optional[Type]] = None,
+                 bound_args: Optional[List[Optional[Type]]] = None,
                  ) -> None:
         if variables is None:
             variables = []
@@ -660,8 +717,7 @@ class CallableType(FunctionLike):
         )
 
     def is_type_obj(self) -> bool:
-        t = self.fallback.type
-        return t is not None and t.is_metaclass()
+        return self.fallback.type.is_metaclass()
 
     def is_concrete_type_obj(self) -> bool:
         return self.is_type_obj() and self.is_classmethod_class
@@ -776,6 +832,23 @@ class CallableType(FunctionLike):
             a.append(tv.id)
         return a
 
+    def __hash__(self) -> int:
+        return hash((self.ret_type, self.is_type_obj(),
+                     self.is_ellipsis_args, self.name,
+                    tuple(self.arg_types), tuple(self.arg_names), tuple(self.arg_kinds)))
+
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, CallableType):
+            return (self.ret_type == other.ret_type and
+                    self.arg_types == other.arg_types and
+                    self.arg_names == other.arg_names and
+                    self.arg_kinds == other.arg_kinds and
+                    self.name == other.name and
+                    self.is_type_obj() == other.is_type_obj() and
+                    self.is_ellipsis_args == other.is_ellipsis_args)
+        else:
+            return NotImplemented
+
     def serialize(self) -> JsonDict:
         # TODO: As an optimization, leave out everything related to
         # generic functions for non-generic functions.
@@ -858,6 +931,14 @@ class Overloaded(FunctionLike):
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_overloaded(self)
 
+    def __hash__(self) -> int:
+        return hash(tuple(self.items()))
+
+    def __eq__(self, other: object) -> bool:
+        if not isinstance(other, Overloaded):
+            return NotImplemented
+        return self.items() == other.items()
+
     def serialize(self) -> JsonDict:
         return {'.class': 'Overloaded',
                 'items': [t.serialize() for t in self.items()],
@@ -899,6 +980,14 @@ class TupleType(Type):
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_tuple_type(self)
 
+    def __hash__(self) -> int:
+        return hash((tuple(self.items), self.fallback))
+
+    def __eq__(self, other: object) -> bool:
+        if not isinstance(other, TupleType):
+            return NotImplemented
+        return self.items == other.items and self.fallback == other.fallback
+
     def serialize(self) -> JsonDict:
         return {'.class': 'TupleType',
                 'items': [t.serialize() for t in self.items],
@@ -913,15 +1002,16 @@ class TupleType(Type):
                          Instance.deserialize(data['fallback']),
                          implicit=data['implicit'])
 
-    def copy_modified(self, *, fallback: Instance = None,
-                      items: List[Type] = None) -> 'TupleType':
+    def copy_modified(self, *, fallback: Optional[Instance] = None,
+                      items: Optional[List[Type]] = None) -> 'TupleType':
         if fallback is None:
             fallback = self.fallback
         if items is None:
             items = self.items
         return TupleType(items, fallback, self.line, self.column)
 
-    def slice(self, begin: int, stride: int, end: int) -> 'TupleType':
+    def slice(self, begin: Optional[int], stride: Optional[int],
+              end: Optional[int]) -> 'TupleType':
         return TupleType(self.items[begin:end:stride], self.fallback,
                          self.line, self.column, self.implicit)
 
@@ -951,6 +1041,21 @@ class TypedDictType(Type):
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_typeddict_type(self)
 
+    def __hash__(self) -> int:
+        return hash((frozenset(self.items.items()), self.fallback,
+                     frozenset(self.required_keys)))
+
+    def __eq__(self, other: object) -> bool:
+        if isinstance(other, TypedDictType):
+            if frozenset(self.items.keys()) != frozenset(other.items.keys()):
+                return False
+            for (_, left_item_type, right_item_type) in self.zip(other):
+                if not left_item_type == right_item_type:
+                    return False
+            return self.fallback == other.fallback and self.required_keys == other.required_keys
+        else:
+            return NotImplemented
+
     def serialize(self) -> JsonDict:
         return {'.class': 'TypedDictType',
                 'items': [[n, t.serialize()] for (n, t) in self.items.items()],
@@ -975,9 +1080,9 @@ class TypedDictType(Type):
         assert self.fallback.type.typeddict_type is not None
         return self.fallback.type.typeddict_type.as_anonymous()
 
-    def copy_modified(self, *, fallback: Instance = None,
-                      item_types: List[Type] = None,
-                      required_keys: Set[str] = None) -> 'TypedDictType':
+    def copy_modified(self, *, fallback: Optional[Instance] = None,
+                      item_types: Optional[List[Type]] = None,
+                      required_keys: Optional[Set[str]] = None) -> 'TypedDictType':
         if fallback is None:
             fallback = self.fallback
         if item_types is None:
@@ -1048,6 +1153,14 @@ class UnionType(Type):
         self.can_be_false = any(item.can_be_false for item in items)
         super().__init__(line, column)
 
+    def __hash__(self) -> int:
+        return hash(frozenset(self.items))
+
+    def __eq__(self, other: object) -> bool:
+        if not isinstance(other, UnionType):
+            return NotImplemented
+        return frozenset(self.items) == frozenset(other.items)
+
     @staticmethod
     def make_union(items: List[Type], line: int = -1, column: int = -1) -> Type:
         if len(items) > 1:
@@ -1162,7 +1275,7 @@ class PartialType(Type):
     inner_types = None  # type: List[Type]
 
     def __init__(self,
-                 type: Optional['mypy.nodes.TypeInfo'],
+                 type: 'Optional[mypy.nodes.TypeInfo]',
                  var: 'mypy.nodes.Var',
                  inner_types: List[Type]) -> None:
         self.type = type
@@ -1227,10 +1340,7 @@ class TypeType(Type):
         type UnionType must be handled through make_normalized static method.
         """
         super().__init__(line, column)
-        if isinstance(item, CallableType) and item.is_type_obj():
-            self.item = item.fallback
-        else:
-            self.item = item
+        self.item = item
 
     @staticmethod
     def make_normalized(item: Type, *, line: int = -1, column: int = -1) -> Type:
@@ -1244,6 +1354,14 @@ class TypeType(Type):
     def accept(self, visitor: 'TypeVisitor[T]') -> T:
         return visitor.visit_type_type(self)
 
+    def __hash__(self) -> int:
+        return hash(self.item)
+
+    def __eq__(self, other: object) -> bool:
+        if not isinstance(other, TypeType):
+            return NotImplemented
+        return self.item == other.item
+
     def serialize(self) -> JsonDict:
         return {'.class': 'TypeType', 'item': self.item.serialize()}
 
@@ -1253,6 +1371,53 @@ class TypeType(Type):
         return TypeType.make_normalized(deserialize_type(data['item']))
 
 
+class ForwardRef(Type):
+    """Class to wrap forward references to other types.
+
+    This is used when a forward reference to an (unanalyzed) synthetic type is found,
+    for example:
+
+        x: A
+        A = TypedDict('A', {'x': int})
+
+    To avoid false positives and crashes in such situations, we first wrap the first
+    occurrence of 'A' in ForwardRef. Then, the wrapped UnboundType is updated in the third
+    pass of semantic analysis and ultimately fixed in the patches after the third pass.
+    So that ForwardRefs are temporary and will be completely replaced with the linked types
+    or Any (to avoid cyclic references) before the type checking stage.
+    """
+    _unbound = None  # type: UnboundType  # The original wrapped type
+    _resolved = None  # type: Optional[Type]  # The resolved forward reference (initially None)
+
+    def __init__(self, unbound: UnboundType) -> None:
+        self._unbound = unbound
+        self._resolved = None
+
+    @property
+    def unbound(self) -> UnboundType:
+        # This is read-only to make it clear that resolution happens through resolve().
+        return self._unbound
+
+    @property
+    def resolved(self) -> Optional[Type]:
+        # Similar to above.
+        return self._resolved
+
+    def resolve(self, resolved: Type) -> None:
+        """Resolve an unbound forward reference to point to a type."""
+        assert self._resolved is None
+        self._resolved = resolved
+
+    def accept(self, visitor: 'TypeVisitor[T]') -> T:
+        return visitor.visit_forwardref_type(self)
+
+    def serialize(self) -> str:
+        name = self.unbound.name
+        # We should never get here since all forward references should be resolved
+        # and removed during semantic analysis.
+        assert False, "Internal error: Unresolved forward reference to {}".format(name)
+
+
 #
 # Visitor-related classes
 #
@@ -1328,6 +1493,9 @@ class TypeVisitor(Generic[T]):
     def visit_type_type(self, t: TypeType) -> T:
         pass
 
+    def visit_forwardref_type(self, t: ForwardRef) -> T:
+        raise RuntimeError('Internal error: unresolved forward reference')
+
 
 class SyntheticTypeVisitor(TypeVisitor[T]):
     """A TypeVisitor that also knows how to visit synthetic AST constructs.
@@ -1430,6 +1598,9 @@ class TypeTranslator(TypeVisitor[Type]):
     def visit_type_type(self, t: TypeType) -> Type:
         return TypeType.make_normalized(t.item.accept(self), line=t.line, column=t.column)
 
+    def visit_forwardref_type(self, t: ForwardRef) -> Type:
+        return t
+
 
 class TypeStrVisitor(SyntheticTypeVisitor[str]):
     """Visitor for pretty-printing types into strings.
@@ -1443,7 +1614,7 @@ class TypeStrVisitor(SyntheticTypeVisitor[str]):
      - Represent the NoneTyp type as None.
     """
 
-    def __init__(self, id_mapper: IdMapper = None) -> None:
+    def __init__(self, id_mapper: Optional[IdMapper] = None) -> None:
         self.id_mapper = id_mapper
 
     def visit_unbound_type(self, t: UnboundType)-> str:
@@ -1585,6 +1756,12 @@ class TypeStrVisitor(SyntheticTypeVisitor[str]):
     def visit_type_type(self, t: TypeType) -> str:
         return 'Type[{}]'.format(t.item.accept(self))
 
+    def visit_forwardref_type(self, t: ForwardRef) -> str:
+        if t.resolved:
+            return '~{}'.format(t.resolved.accept(self))
+        else:
+            return '~{}'.format(t.unbound.accept(self))
+
     def list_str(self, a: List[Type]) -> str:
         """Convert items of an array to strings (pretty-print types)
         and join the results with commas.
@@ -1664,6 +1841,12 @@ class TypeQuery(SyntheticTypeVisitor[T]):
     def visit_type_type(self, t: TypeType) -> T:
         return t.item.accept(self)
 
+    def visit_forwardref_type(self, t: ForwardRef) -> T:
+        if t.resolved:
+            return t.resolved.accept(self)
+        else:
+            return t.unbound.accept(self)
+
     def visit_ellipsis_type(self, t: EllipsisType) -> T:
         return self.strategy([])
 
@@ -1762,20 +1945,20 @@ def function_type(func: mypy.nodes.FuncBase, fallback: Instance) -> FunctionLike
         # Implicit type signature with dynamic types.
         # Overloaded functions always have a signature, so func must be an ordinary function.
         assert isinstance(func, mypy.nodes.FuncItem), str(func)
-        return callable_type(cast(mypy.nodes.FuncItem, func), fallback)
+        return callable_type(func, fallback)
 
 
 def callable_type(fdef: mypy.nodes.FuncItem, fallback: Instance,
-                  ret_type: Type = None) -> CallableType:
+                  ret_type: Optional[Type] = None) -> CallableType:
     name = fdef.name()
     if name:
         name = '"{}"'.format(name)
 
     return CallableType(
-        [AnyType()] * len(fdef.arg_names),
+        [AnyType(TypeOfAny.unannotated)] * len(fdef.arg_names),
         fdef.arg_kinds,
         [None if argument_elide_name(n) else n for n in fdef.arg_names],
-        ret_type or AnyType(),
+        ret_type or AnyType(TypeOfAny.unannotated),
         fallback,
         name,
         implicit=True,
